import java.io.*;
import java.net.*;
import java.util.*;

/**
Euler proof mechanism.
  @version 23
  @author Jos De Roo
                                                                                                    */
public class Euler {
  Euler subject = null;    // RDF subject
  String predicate = null; // RDF predicate
  Euler object = null;     // RDF object
  String varname = null;   // variable name
  int varid = 0;           // variable id
  boolean bound = false;   // variable binding state
  boolean not = false;     // predicate not state
  Euler ref = null;        // to be dereferenced  
  Euler next = null;       // to construct a rule
  Euler near = null;       // to chain assertions
  static final String NS = "http://www.agfa.com/w3c/euler#";
  static int varmax = 0;

  Euler(String nt, StringTokenizer st, Hashtable vt) {
    if (!st.hasMoreTokens()) return;
    while (nt.equals(" ") || nt.equals("\n")) nt = st.nextToken();
    boolean b = nt.equals("[");
    if (b) {
      nt = st.nextToken();
      while (nt.equals(" ") || nt.equals("\n")) nt = st.nextToken();
      subject = new Euler(nt, st, vt);
      nt = st.nextToken();
      while (nt.equals(" ") || nt.equals("\n")) nt = st.nextToken();
    }
    predicate = nt;
    if (predicate.startsWith("data:?")) {
      varname = predicate;
      Euler t = (Euler)vt.get(predicate);
      if (t == null) {
        varid = vt.size();
        vt.put(predicate, this);
        if (varid + 1 > varmax) varmax = varid + 1;
      }
      else varid = t.varid;
    }
    else bound = true;
    if (predicate.endsWith("-not")) {
      predicate = predicate.substring(0, predicate.lastIndexOf("-not"));
      not = true;
    }
    if (b) {
      nt = st.nextToken();
      while (nt.equals(" ") || nt.equals("\n")) nt = st.nextToken();
      object = new Euler(nt, st, vt);
      nt = st.nextToken();
      while (nt.equals(" ") || nt.equals("\n")) nt = st.nextToken();
    }
  }

/**
  constructs a proof engine
                                                                                                    */
  public Euler() {
  }

/**
  asserts all facts and rules acquired from the RDF URI in that proof engine.
  @param uri of the RDF resource
                                                                                                    */
  public void assert(String uri) {
    String rk = fromStatement(fromWeb(uri));
    StringTokenizer st = new StringTokenizer(rk, "[] \n", true);
    Hashtable vt = new Hashtable();
    Euler lr = this;
    while (lr.near != null) lr = lr.near;
    while (st.hasMoreTokens()) {
      String nt = st.nextToken();
      lr.near = new Euler(nt, st, vt);
      Euler llr = lr.near;
      while (llr.predicate.equals(NS + "if") || llr.predicate.equals(NS + "and")) {
        if (llr.object.predicate.equals(NS + "and")) llr.subject.next = llr.object.subject;
        else llr.subject.next = llr.object;
        llr = llr.object;
      }
      lr = lr.near;
    }
  }

/**
  proofs a conjunction with that proof engine
  @param uri of the RDF resource
  @return proof
                                                                                                    */
  public String proof(String uri) {
    String cj = fromStatement(fromWeb(uri));
    StringTokenizer st = new StringTokenizer(cj, "[] \n", true);
    Hashtable vt = new Hashtable();
    Euler goal = new Euler(st.nextToken(), st, vt);
    Euler lr = goal;
    while (lr.predicate.equals(NS + "and")) {
      if (lr.object.predicate.equals(NS + "and")) lr.subject.next = lr.object.subject;
      else lr.subject.next = lr.object;
      lr = lr.object;
    }
    if (goal.predicate.equals(NS + "and")) goal = goal.subject;
    Euler[] vars = new Euler[varmax];
    Euler terms = null;
    Euler termsb = null;
    Euler terms1 = null;
    Euler terms2 = null;
    Stack stack = new Stack();
    String predicate = null;
    boolean not = false;
    int step = 0;
    String proof = "";
    String pp = "";
    long t = System.currentTimeMillis();
    while (true) {
      if (goal == null) {
        t = System.currentTimeMillis() - t + 1;
        System.err.println("<proof steps=\"" + step + "\" steps/sec=\"" + step*1000L/t + "\"/>");
        if (!proof.equals("")) {
          st = new StringTokenizer(proof, "[] \n", true);
          vt = new Hashtable();
          proof = new Euler(st.nextToken(), st, vt).toStatement(null, "  ");
        }
        return "<r:RDF xmlns:r='http://www.w3.org/1999/02/22-rdf-syntax-ns#'>\n" + proof + "</r:RDF>";
      }
      step++;
      predicate = goal.predicate;
      not = goal.not;
      if (predicate == null) pp = "[" + pp + " " + NS + "thus\n " + goal.subject + "]";
      else if (predicate.equals(NS + "deref")) {
        st = new StringTokenizer(goal.subject.toString(), "=&;", true);
        String u = "";
        vt = new Hashtable();
        while (st.hasMoreTokens()) {
          String nt = st.nextToken();
          if (nt.equals("="))
            u = u + "=" + URLEncoder.encode(new Euler(st.nextToken(), st, vt).eval(vars).toString());
          else if (nt.equals("&")) u = u + URLEncoder.encode("&");
          else u = u + nt;
        }
        String rk = fromStatement(fromWeb(u));
        st = new StringTokenizer(rk, "[] \n", true);
        Euler el = new Euler(st.nextToken(), st, vt);
        if (goal.object.unify(el, stack)) {
          u = "[" + u + " " + predicate + " " + goal.object + "]";
          if (pp.equals("")) pp = u;
          else pp = "[" + pp + " e:and\n" + u + "]";
        }
        else not = !not;
      }
      else {
        if (terms == null) {
          lr = this.near;
          while (lr != null) {
            terms = lr;
            if (terms.predicate.equals(predicate)) break;
            if (terms.predicate.equals(NS + "if") && terms.subject.predicate.equals(predicate)) {
              terms = terms.subject;
              break;
            }
            else lr = lr.near;
            terms = null;
          }
        }
        if (terms != null) {
          Euler n = lr.near;
          termsb = null;
          while (n != null) {
            termsb = n;
            if (termsb.predicate.equals(predicate)) break;
            if (termsb.predicate.equals(NS + "if") && termsb.subject.predicate.equals(predicate)) {
              termsb = termsb.subject;
              break;
            }
            else n = n.near;
            termsb = null;
          }
          if (termsb != null) {
            stack.push(goal);
            stack.push(termsb);
            stack.push(n);
            stack.push(pp);
          }
          for (int i = 0; i < varmax; i++) vars[i] = null;
          if (terms.eval(vars).unify(goal, stack)) {
            predicate = goal.eval(vars).toString();
            Euler ep = goal.next;
            while (ep != null) {
              if (ep.predicate == null && ep.subject.subject.bound && ep.subject.object.bound
                  && ep.subject.toString().equals(predicate)) break;
              ep = ep.next;
            }
            if (ep != null) not = true;
            else {
              terms = terms.next;
              if (terms != null) {
                terms2 = new Euler();
                terms1 = terms2;
                while (terms != null) {
                  terms1.next = terms.eval(vars);
                  terms1 = terms1.next;
                  terms = terms.next;
                }
                terms1.next = new Euler();
                terms1 = terms1.next;
                terms1.subject = goal;
                terms1.next = goal.next;
                goal = terms2;
              }
              else {
                if (pp.equals("")) pp = goal.toString();
                else pp = "[" + pp + " " + NS + "and\n" + goal.toString() + "]";
              }
            }
          }
          else not = !not;
        }
        else not = !not;
      }
      goal = goal.next;
      if (not || goal == null) {
        if (goal == null && !pp.equals("")) {
          if (proof.equals("")) proof = pp;
          else proof = "[" + proof + "\n\n" + NS + "more\n" + pp + "]";
        }
        goal = null;
        while (!stack.empty()) {
          Object b = stack.pop();
          if (b instanceof Euler) ((Euler)b).bound = false;
          else if (b instanceof String) {
            pp = (String)b;
            lr = (Euler)stack.pop();
            terms = (Euler)stack.pop();
            goal = (Euler)stack.pop();
            break;
          }
        }
      }
    }
  }

  final Euler eval(Euler[] v) {
    if (bound) {
      if (ref != null) return ref.eval(v);
      Euler t = new Euler();
      t.predicate = predicate;
      if (subject != null) t.subject = subject.eval(v);
      if (object != null) t.object = object.eval(v);
      t.bound = true;
      return t;
    }
    if (v[varid] == null) {
      v[varid] = new Euler();
      v[varid].predicate = predicate;
      v[varid].varname = varname;
      v[varid].varid = varid;
    }
    return v[varid];
  }

  final boolean unify(Euler t, Stack s) {
    if (bound && ref != null) return ref.unify(t, s);
    if (t.bound && t.ref != null) return unify(t.ref, s);
    if (bound && t.bound) {
      if (!predicate.equals(t.predicate)) return false;
      if (subject != null && !subject.unify(t.subject, s)) return false;
      if (object != null && !object.unify(t.object, s)) return false;
      return true;
    }
    if (bound) return t.unify(this, s);
    bound = true;
    ref = t;
    s.push(this);
    return true;
  }

  final String fromWeb(String uri) {
    String rk = "";
    String rl = "";
    try {
      URL r = new URL(uri);
      BufferedReader br = new BufferedReader(new InputStreamReader(r.openStream()));
      while ((rl = br.readLine()) != null) rk = rk + rl + "\n";
      return rk;
    }
    catch (Exception e) {
      System.err.println(e);
      return null;
    }
  }

  final String toStatement(String kind, String indent) {
    if (ref != null) return ref.toStatement(kind, indent);
    if (subject == null && object == null)
      return indent+"<"+kind+" r:resource=\""+(bound?predicate:varname)+"\"/>\n";
    String tab = indent+(kind==null?"":"  ");
    return (kind==null?"":indent+"<"+kind+">\n")+tab+"<r:Statement>\n"+
           subject.toStatement("r:subject", tab+"  ")+
           tab+"  <r:predicate r:resource=\""+(bound?predicate:varname)+(not?"-not":"")+"\"/>\n"+
           object.toStatement("r:object", tab+"  ")+
           tab+"</r:Statement>\n"+(kind==null?"":indent+"</"+kind+">\n");
  }

  final String fromStatement(String s) {
    StringTokenizer st = new StringTokenizer(s, "<>\"", true);
    String nt = st.nextToken();
    String rs = "";
    while (st.hasMoreTokens()) {
      if (nt.endsWith(":Statement")) rs = rs + (!nt.startsWith("/")?"[":"]");
      else if (!rs.equals("") && nt.equals("\"")) {
        rs = rs + " " + st.nextToken() + " ";
        st.nextToken();
      }
      nt = st.nextToken();
    }
    return rs;
  }

  public final String toString() {
    if (ref != null) return ref.toString();
    if (subject == null && object == null) return bound?predicate:varname;
    return "[" + subject + " " + (bound?predicate:varname) + (not?"-not":"") + " " + object + "]";
  }
}
